from abc import ABC, abstractmethod
import time
from typing import Optional, TypeVar, Type, Generic

from pydantic import BaseModel
from toolfuse import Tool

from .models import V1Device, V1Lock, V1DeviceType

D = TypeVar("D", bound="Device")
C = TypeVar("C", bound="BaseModel")
P = TypeVar("P", bound="BaseModel")


class ReactComponent:
    """A react component for a device"""

    source: str
    server_uri: str
    token: Optional[str] = None


class Device(Generic[C, D, P], Tool, ABC):
    """An agent device"""

    @classmethod
    @abstractmethod
    def connect(cls, config: C) -> D:
        """Connect to a device from a configuration

        Args:
            config (C): Config

        Returns:
            D: The device
        """
        pass

    @abstractmethod
    def disconnect(self) -> None:
        """Disconnect from the device"""
        pass

    def name(self) -> Optional[str]:
        """Name of the device"""
        pass

    @classmethod
    @abstractmethod
    def ensure(cls, name: str, config: P) -> D:
        """Ensure device infrastructure exists

        Args:
            name (str): Name of the device
            config (P): Provisioner configuration

        Returns:
            D: the device
        """
        pass

    @classmethod
    @abstractmethod
    def create(cls, name: str, config: P) -> D:
        """Create device infrastructure

        Args:
            name (str): Name of the device
            config (P): Provisioner configuration

        Returns:
            D: The device
        """
        pass

    @classmethod
    @abstractmethod
    def react_component(cls) -> Optional[ReactComponent]:
        """React component for the device

        Returns:
            Optional[ReactComponent]: React component
        """
        pass

    @abstractmethod
    def view(self, background: bool = False) -> None:
        """View the device in the browser

        Args:
            background (bool, optional): Whether to run in the background. Defaults to False.
        """
        pass

    @abstractmethod
    def connect_config(self) -> C:
        """Connect configuration

        Returns:
            C: Connect configuration for this device
        """
        pass

    @classmethod
    @abstractmethod
    def connect_config_type(cls) -> Type[C]:
        """Type of connect configuration

        Returns:
            Type[C]: Type of connect configuration
        """
        pass

    @classmethod
    @abstractmethod
    def provision_config_type(cls) -> Type[P]:
        """Type of provision configuration

        Returns:
            Type[P]: Type of provisioner configuration
        """
        pass

    def get_lock(self) -> Optional[V1Lock]:
        """Get the lock for the device"""
        if not hasattr(self, "_lock"):
            return None
        return self._lock

    def is_locked(self) -> bool:
        """Check if the device is locked"""
        if not hasattr(self, "_lock"):
            return False
        if self._lock is None:
            return False
        return self._lock.locked

    def unlock(self) -> None:
        """Unlock the device"""
        self._lock = None

    def lock(self, owner_id: str, expires_in: int = 600) -> None:
        """Lock the device"""
        if hasattr(self, "_lock"):
            if self._lock and self._lock.locked:
                raise RuntimeError("Device is already locked")
        self._lock = V1Lock(
            owner_id=owner_id,
            created=time.time(),
            expires=time.time() + expires_in,
            locked=True,
        )

    def to_v1(self) -> V1Device:
        """Convert the device to v1 server transport

        Returns:
            V1Device: A v1 device
        """
        config_type = self.connect_config_type()
        config_instance = self.connect_config()
        if not isinstance(config_instance, config_type):
            raise TypeError(
                f"Expected config instance of type {config_type.__name__}, but got {type(config_instance).__name__}"
            )
        parametrized_device_model = V1Device[config_type](
            name=self.name(), type=self.type(), config=config_instance
        )
        return parametrized_device_model

    @classmethod
    def serve(cls) -> None:
        pass


class MultiDevice(Device):
    """A device of multiple devices"""

    pass



from typing import Optional, Type, List

from pydantic import BaseModel

from devicebay import Device, action, observation, ReactComponent


class FSConnectConfig(BaseModel):
    """Connect configuration for a filesystem"""

    path: str


class FSProvisionConig(BaseModel):
    """Provision configuration for a filesystem"""

    path: str


class FileInfo(BaseModel):
    """File information"""

    name: str
    type: int
    created: float
    updated: float
    size: int


class FileSystem(Device[FSConnectConfig, "FileSystem", FSProvisionConig]):
    """A filesystem device"""

    def __init__(self, root_path: str) -> None:
        super().__init__()
        self._root_path = root_path
        self._config: FSConnectConfig = FSConnectConfig(path=root_path)

    @classmethod
    def connect(cls, config: FSConnectConfig) -> "FileSystem":
        """Connect to a device from a configuration

        Args:
            config (C): Config

        Returns:
            D: The device
        """
        return cls(config.path)

    def disconnect(self) -> None:
        """Disconnect from the device"""
        pass

    @classmethod
    def ensure(cls, name: str, config: FSProvisionConig) -> "FileSystem":
        """Ensure device infrastructure exists

        Args:
            name (str): Name of the device
            config (P): Provisioner configuration

        Returns:
            D: the device
        """
        return cls(config.path)

    @classmethod
    def create(cls, name: str, config: FSProvisionConig) -> "FileSystem":
        """Create device infrastructure

        Args:
            name (str): Name of the device
            config (P): Provisioner configuration

        Returns:
            D: The device
        """
        return cls(config.path)

    @classmethod
    def react_component(cls) -> Optional[ReactComponent]:
        """React component for the device

        Returns:
            ReactComponent: React component
        """
        return None

    def view(self, background: bool = False) -> None:
        """View the device in the browser

        Args:
            background (bool, optional): Whether to run in the background. Defaults to False.
        """
        pass

    def connect_config(self) -> FSConnectConfig:
        """Connect configuration

        Returns:
            FSConnectConfg: Connect configuration for this device
        """
        return FSConnectConfig(path=self._config.path)

    @classmethod
    def connect_config_type(cls) -> Type[FSConnectConfig]:
        """Type of connect configuration

        Returns:
            Type[FSConnectConfig]: Type of connect configuration
        """
        return FSConnectConfig

    @classmethod
    def provision_config_type(cls) -> Type[FSProvisionConig]:
        """Type of provision configuration

        Returns:
            Type[P]: Type of provisioner configuration
        """
        return FSProvisionConig

    @action
    def create_file(self, path: str, content: str) -> None:
        """Create a file

        Args:
            path (str): Path relative to root path
            content (str): Content of the file
        """
        with open(f"{self._root_path}/{path}", "w") as f:
            f.write(content)

    @observation
    def read_file(self, path: str) -> str:
        """Read a file

        Args:
            path (str): Path relative to root path

        Returns:
            str: The file
        """
        with open(f"{self._root_path}/{path}", "r") as f:
            return f.read()

    @action
    def create_dir(self, path: str) -> None:
        """Create a directory

        Args:
            path (str): Path relative to root path
        """
        import os

        os.makedirs(f"{self._root_path}/{path}")

    @observation
    def list_dir(self, path: str) -> List[FileInfo]:
        """List a directory

        Args:
            path (str): Path relative to root path

        Returns:
            List[FileInfo]: The file info
        """
        import os
        import stat
        from datetime import datetime

        files = []
        for f in os.listdir(f"{self._root_path}/{path}"):
            f_path = f"{self._root_path}/{path}/{f}"
            st = os.stat(f_path)
            files.append(
                FileInfo(
                    name=f,
                    type=stat.S_IFMT(st.st_mode),
                    created=datetime.fromtimestamp(st.st_ctime).timestamp(),
                    updated=datetime.fromtimestamp(st.st_mtime).timestamp(),
                    size=st.st_size,
                )
            )
        return files

    @action
    def delete_file(self, path: str) -> None:
        """Delete a file

        Args:
            path (str): Path relative to root path
        """
        import os

        os.remove(f"{self._root_path}/{path}")

    @action
    def delete_dir(self, path: str) -> None:
        """Delete a directory

        Args:
            path (str): Path relative to root path
        """
        import os

        os.rmdir(f"{self._root_path}/{path}")

    @action
    def append_file(self, path: str, content: str) -> None:
        """Append to a file

        Args:
            path (str): Path relative to root path
            content (str): Content to append
        """
        with open(f"{self._root_path}/{path}", "a") as f:
            f.write(content)

    @action
    def get_diff(self, path: str, content: str) -> str:
        """Get diff of a file

        Args:
            path (str): Path relative to root path
            content (str): Content to diff against

        Returns:
            str: The diff in unified diff format
        """
        import difflib

        # Correcting the splitlines usage by explicitly using `keepends=True`
        with open(f"{self._root_path}/{path}", "r") as f:
            old = f.read()
            # Splitting lines and keeping line breaks
            diff = difflib.unified_diff(
                old.splitlines(keepends=True),
                content.splitlines(keepends=True),
                fromfile=path,
                tofile=path,
            )
        return "\n".join(diff)

    @action
    def overwrite_file(self, path: str, content: str) -> None:
        """Overwrite a file

        Args:
            path (str): Path relative to root path
            content (str): Content to overwrite with
        """
        with open(f"{self._root_path}/{path}", "w") as f:
            f.write(content)



import uuid
import time

from sqlalchemy import Column, String, Boolean, Float, Integer
from sqlalchemy.orm import declarative_base

Base = declarative_base()


class UserRecord(Base):
    __tablename__ = "users"

    email = Column(String, unique=True, index=True, primary_key=True)
    display_name = Column(String)
    handle = Column(String)
    picture = Column(String)
    created = Column(Integer)
    updated = Column(Integer)


class DeviceTypeRecord(Base):
    __tablename__ = "device_types"

    id = Column(String, primary_key=True, default=lambda: str(uuid.uuid4()))
    name = Column(String, unique=True, index=True)
    description = Column(String)
    image = Column(String)
    versions = Column(String)
    env_opts = Column(String)
    supported_runtimes = Column(String)
    owner_id = Column(String)
    public = Column(Boolean)
    icon = Column(String)
    created = Column(Float, default=time.time)
    updated = Column(Float, default=time.time)
    mem_request = Column(String)
    mem_limit = Column(String)
    cpu_request = Column(String)
    cpu_limit = Column(String)
    gpu_mem = Column(String)
    llm_providers = Column(String, nullable=True)



from typing import Optional, Type

from pydantic import BaseModel
from playwright.sync_api import sync_playwright, Browser, Page

from devicebay import Device, action, observation, ReactComponent


class PlaywrightConnectConfig(BaseModel):
    """Connect configuration for Playwright"""

    browser_type: str = "chromium"
    headless: bool = True


class PlaywrightProvisionConfig(BaseModel):
    """Provision configuration for Playwright"""

    pass


class Playwright(
    Device[PlaywrightConnectConfig, "Playwright", PlaywrightProvisionConfig]
):
    """A Playwright device"""

    def __init__(
        self,
        config: PlaywrightConnectConfig,
    ) -> None:
        super().__init__()
        self._config = config
        self._playwright = sync_playwright().start()
        self._browser: Browser = getattr(self._playwright, config.browser_type).launch(
            headless=config.headless
        )
        self._page: Page = self._browser.new_page()

    @classmethod
    def connect(cls, config: PlaywrightConnectConfig) -> "Playwright":
        """Connect to a device from a configuration

        Args:
            config (PlaywrightConnectConfig): Config

        Returns:
            Playwright: The device
        """
        return cls(config)

    def disconnect(self) -> None:
        """Disconnect from the device"""
        self._browser.close()
        self._playwright.stop()

    @classmethod
    def ensure(cls, name: str, config: PlaywrightProvisionConfig) -> "Playwright":
        """Ensure device infrastructure exists

        Args:
            name (str): Name of the device
            config (PlaywrightProvisionConfig): Provisioner configuration

        Returns:
            Playwright: The device
        """
        return cls(PlaywrightConnectConfig())

    @classmethod
    def create(cls, name: str, config: PlaywrightProvisionConfig) -> "Playwright":
        """Create device infrastructure

        Args:
            name (str): Name of the device
            config (PlaywrightProvisionConfig): Provisioner configuration

        Returns:
            Playwright: The device
        """
        return cls(PlaywrightConnectConfig())

    @classmethod
    def react_component(cls) -> Optional[ReactComponent]:
        """React component for the device

        Returns:
            ReactComponent: React component
        """
        return None

    def view(self, background: bool = False) -> None:
        """View the device in the browser

        Args:
            background (bool, optional): Whether to run in the background. Defaults to False.
        """
        pass

    def connect_config(self) -> PlaywrightConnectConfig:
        """Connect configuration

        Returns:
            PlaywrightConnectConfig: Connect configuration for this device
        """
        return self._config

    @classmethod
    def connect_config_type(cls) -> Type[PlaywrightConnectConfig]:
        """Type of connect configuration

        Returns:
            Type[PlaywrightConnectConfig]: Type of connect configuration
        """
        return PlaywrightConnectConfig

    @classmethod
    def provision_config_type(cls) -> Type[PlaywrightProvisionConfig]:
        """Type of provision configuration

        Returns:
            Type[PlaywrightProvisionConfig]: Type of provisioner configuration
        """
        return PlaywrightProvisionConfig

    @action
    def navigate(self, url: str) -> None:
        """Navigate to a URL

        Args:
            url (str): URL to navigate to
        """
        self._page.goto(url)

    @observation
    def get_url(self) -> str:
        """Get the current URL

        Returns:
            str: The current URL
        """
        return self._page.url

    @action
    def click(self, selector: str) -> None:
        """Click on an element

        Args:
            selector (str): CSS selector of the element to click
        """
        self._page.click(selector)

    @action
    def type(self, selector: str, text: str) -> None:
        """Type text into an element

        Args:
            selector (str): CSS selector of the element to type into
            text (str): Text to type
        """
        self._page.fill(selector, text)

    @observation
    def get_text(self, selector: str) -> str:
        """Get the text of an element

        Args:
            selector (str): CSS selector of the element

        Returns:
            str: The text of the element
        """
        return self._page.inner_text(selector)

    @observation
    def get_page_source(self) -> str:
        """Get the page source

        Returns:
            str: The page source
        """
        return self._page.content()

    @action
    def press_key(self, selector: str, key: str) -> None:
        """Press a key on an element

        Args:
            selector (str): CSS selector of the element
            key (str): Key to press
        """
        self._page.press(selector, key)

    @action
    def hover(self, selector: str) -> None:
        """Hover over an element

        Args:
            selector (str): CSS selector of the element to hover over
        """
        self._page.hover(selector)

    @observation
    def get_attribute(self, selector: str, name: str) -> Optional[str]:
        """Get an attribute of an element

        Args:
            selector (str): CSS selector of the element
            name (str): Name of the attribute

        Returns:
            str: The attribute value
        """
        return self._page.get_attribute(selector, name)



from typing import List, Optional, Dict
import uuid
import time
import json

from sqlalchemy import or_

from .db.models import DeviceTypeRecord
from .db.conn import WithDB
from .models import (
    V1EnvVarOpt,
    V1LLMProviders,
    V1DeviceTypeFile,
    V1LLMProviders,
)


class DeviceType(WithDB):
    """A type of device"""

    def __init__(
        self,
        name: str,
        description: str,
        image: str,
        versions: Dict[str, str],
        supported_runtimes: List[str],
        owner_id: Optional[str] = None,
        env_opts: List[V1EnvVarOpt] = [],
        public: bool = False,
        icon: Optional[str] = None,
        mem_request: Optional[str] = "500m",
        mem_limit: Optional[str] = "2gi",
        cpu_request: Optional[str] = "1",
        cpu_limit: Optional[str] = "4",
        gpu_mem: Optional[str] = None,
        llm_providers: Optional[V1LLMProviders] = None,
    ):
        self.id = str(uuid.uuid4())
        self.name = name
        self.description = description
        self.image = image
        self.versions = versions
        self.supported_runtimes = supported_runtimes
        self.owner_id = owner_id
        self.env_opts = env_opts
        self.public = public
        self.icon = icon
        self.mem_request = mem_request
        self.mem_limit = mem_limit
        self.cpu_request = cpu_request
        self.cpu_limit = cpu_limit
        self.gpu_mem = gpu_mem
        self.created = time.time()
        self.updated = time.time()
        self.llm_providers: Optional[V1LLMProviders] = llm_providers
        self.save()

    def to_schema(self) -> V1DeviceTypeFile:
        return V1DeviceTypeFile(
            id=self.id,
            name=self.name,
            description=self.description,
            image=self.image,
            versions=self.versions,
            env_opts=self.env_opts,
            supported_runtimes=self.supported_runtimes,
            created=self.created,
            updated=self.updated,
            public=self.public,
            icon=self.icon,
            mem_request=self.mem_request,
            mem_limit=self.mem_limit,
            cpu_request=self.cpu_request,
            cpu_limit=self.cpu_limit,
            gpu_mem=self.gpu_mem,
            llm_providers=self.llm_providers,
            owner_id=self.owner_id,
        )

    @classmethod
    def from_schema(cls, schema: V1DeviceTypeFile) -> "DeviceType":
        obj = cls.__new__(cls)
        obj.id = schema.id
        obj.name = schema.name
        obj.owner_id = schema.owner_id
        obj.description = schema.description
        obj.image = schema.image
        obj.env_opts = schema.env_opts
        obj.supported_runtimes = schema.supported_runtimes
        obj.created = schema.created
        obj.updated = schema.updated
        obj.public = schema.public
        obj.icon = schema.icon
        obj.mem_limit = schema.mem_limit
        obj.mem_request = schema.mem_request
        obj.cpu_limit = schema.cpu_limit
        obj.cpu_request = schema.cpu_request
        obj.gpu_mem = schema.gpu_mem
        obj.versions = schema.versions
        obj.llm_providers = schema.llm_providers
        return obj

    def to_record(self) -> DeviceTypeRecord:
        versions = json.dumps(self.versions)
        llm_providers = None
        if self.llm_providers:
            llm_providers = json.dumps(self.llm_providers.model_dump())

        return DeviceTypeRecord(
            id=self.id,
            name=self.name,
            description=self.description,
            image=self.image,
            versions=versions,
            env_opts=json.dumps([opt.model_dump() for opt in self.env_opts]),
            supported_runtimes=json.dumps(self.supported_runtimes),
            created=self.created,
            updated=self.updated,
            owner_id=self.owner_id,
            public=self.public,
            icon=self.icon,
            mem_limit=self.mem_limit,
            mem_request=self.mem_request,
            cpu_limit=self.cpu_limit,
            cpu_request=self.cpu_request,
            gpu_mem=self.gpu_mem,
            llm_providers=llm_providers,
        )

    @classmethod
    def from_record(cls, record: DeviceTypeRecord) -> "DeviceType":
        versions = {}
        if len(str(record.versions)) != 0:
            versions = json.loads(str(record.versions))

        llm_providers = None
        if len(str(record.llm_providers)) != 0:
            llm_providers = V1LLMProviders(**json.loads(str(record.llm_providers)))

        obj = cls.__new__(cls)
        obj.id = record.id
        obj.name = record.name
        obj.description = record.description
        obj.image = record.image
        obj.versions = versions
        obj.env_opts = [V1EnvVarOpt(**opt) for opt in json.loads(str(record.env_opts))]
        obj.supported_runtimes = json.loads(str(record.supported_runtimes))
        obj.created = record.created
        obj.updated = record.updated
        obj.owner_id = record.owner_id
        obj.public = record.public
        obj.icon = record.icon
        obj.mem_limit = record.mem_limit
        obj.mem_request = record.mem_request
        obj.cpu_limit = record.cpu_limit
        obj.cpu_request = record.cpu_request
        obj.gpu_mem = record.gpu_mem
        obj.llm_providers = llm_providers
        return obj

    def save(self) -> None:
        for session in self.get_db():
            if session:
                record = self.to_record()
                session.merge(record)
                session.commit()

    @classmethod
    def find(cls, **kwargs) -> List["DeviceType"]:
        for session in cls.get_db():
            records = session.query(DeviceTypeRecord).filter_by(**kwargs).all()
            return [cls.from_record(record) for record in records]

        return []

    @classmethod
    def find_for_user(
        cls, user_id: str, name: Optional[str] = None
    ) -> List["DeviceType"]:
        for session in cls.get_db():
            # Base query
            query = session.query(DeviceTypeRecord).filter(
                or_(
                    DeviceTypeRecord.owner_id == user_id,
                    DeviceTypeRecord.public == True,
                )
            )

            # Conditionally add name filter if name is provided
            if name is not None:
                query = query.filter(DeviceTypeRecord.name == name)

            records = query.all()
            return [cls.from_record(record) for record in records]
        return []

    @classmethod
    def delete(cls, id: str, owner_id: str) -> None:
        for session in cls.get_db():
            if session:
                record = (
                    session.query(DeviceTypeRecord)
                    .filter_by(id=id, owner_id=owner_id)
                    .first()
                )
                if record:
                    session.delete(record)
                    session.commit()

    def update(self, model: V1DeviceTypeFile) -> None:
        """
        Updates the current DeviceType instance with values from an DeviceTypeModel instance.
        """
        # Track if any updates are needed
        updated = False

        # Compare and update fields
        if self.name != model.name:
            self.name = model.name
            updated = True

        if self.description != model.description:
            self.description = model.description
            updated = True

        if self.image != model.image:
            self.image = model.image
            updated = True

        if self.versions != model.versions:
            self.versions = model.versions
            updated = True

        if self.env_opts != model.env_opts:
            self.env_opts = model.env_opts
            updated = True

        if self.supported_runtimes != model.supported_runtimes:
            self.supported_runtimes = model.supported_runtimes
            updated = True

        if self.public != model.public:
            self.public = model.public
            updated = True

        if self.icon != model.icon:
            self.icon = model.icon
            updated = True

        if self.mem_request != model.mem_request:
            self.mem_request = model.mem_request
            updated = True

        if self.mem_limit != model.mem_limit:
            self.mem_limit = model.mem_limit
            updated = True

        if self.cpu_request != model.cpu_request:
            self.cpu_request = model.cpu_request
            updated = True

        if self.cpu_limit != model.cpu_limit:
            self.cpu_limit = model.cpu_limit
            updated = True

        if self.gpu_mem != model.gpu_mem:
            self.gpu_mem = model.gpu_mem
            updated = True

        if self.llm_providers != model.llm_providers:
            self.llm_providers = model.llm_providers
            updated = True

        # If anything was updated, set the updated timestamp and save changes
        if updated:
            self.updated = time.time()
            self.save()



from typing import Optional, Dict, List, TypeVar, Generic
from pydantic import BaseModel

ConnectConfigType = TypeVar("ConnectConfigType")
ProvisionConfigType = TypeVar("ProvisionConfigType")


class V1Lock(BaseModel):
    owner_id: str
    created: float
    expires: float
    locked: bool
    reason: Optional[str] = None


class V1Device(BaseModel, Generic[ConnectConfigType]):
    type: str
    name: Optional[str] = None
    config: Optional[ConnectConfigType] = None


class V1DeviceType(BaseModel, Generic[ProvisionConfigType]):
    name: str
    config: Optional[ProvisionConfigType] = None


class V1EnvVarOpt(BaseModel):
    name: str
    description: Optional[str] = None
    required: bool = False
    default: Optional[str] = None
    secret: bool = False
    options: List[str] = []


class V1LLMProviders(BaseModel):
    preference: List[str] = []


class V1DeviceTypeFile(BaseModel):
    id: Optional[str] = None
    name: str
    owner_id: Optional[str] = None
    description: str
    image: Optional[str] = None
    versions: Optional[Dict[str, str]] = None
    env_opts: List[V1EnvVarOpt] = []
    supported_runtimes: List[str] = []
    created: Optional[float] = None
    updated: Optional[float] = None
    public: bool = False
    icon: Optional[str] = None
    mem_request: Optional[str] = "500m"
    mem_limit: Optional[str] = "2gi"
    cpu_request: Optional[str] = "1"
    cpu_limit: Optional[str] = "4"
    gpu_mem: Optional[str] = None
    llm_providers: Optional[V1LLMProviders] = None


class V1DeviceTypeFiles(BaseModel):
    types: List[V1DeviceTypeFile]



from typing import Optional, List, Dict

from pydantic import BaseModel

from github import Github, PullRequest, Issue
from github.Issue import Issue
from github.Repository import Repository
from github.PullRequest import PullRequest

from devicebay import Device, action, observation


class GitHubConnectConfig(BaseModel):
    """Connect configuration for GitHub"""

    access_token: str


class GitHubProvisionConfig(BaseModel):
    """Provision configuration for GitHub"""

    repo: str


class GitHub(Device[GitHubConnectConfig, "GitHub", GitHubProvisionConfig]):
    """A GitHub device"""

    def __init__(self, config: GitHubConnectConfig) -> None:
        super().__init__()
        self._config = config
        self._github = Github(config.access_token)

    @observation
    def get_repository(self, name: str) -> Repository:
        """Get a repository by name

        Args:
            name (str): Name of the repository

        Returns:
            Repository: The repository
        """
        return self._github.get_user().get_repo(name)

    @action
    def create_pull_request(
        self, repo_name: str, title: str, head: str, base: str, body: str = ""
    ) -> None:
        """Create a pull request

        Args:
            repo_name (str): Name of the repository
            title (str): Title of the pull request
            head (str): Branch name or commit SHA of the head
            base (str): Branch name of the base
            body (str, optional): Body of the pull request. Defaults to "".
        """
        repo: Repository = self.get_repository(repo_name)
        repo.create_pull(title=title, body=body, head=head, base=base)

    @observation
    def get_pull_request(self, repo_name: str, pr_number: int) -> PullRequest:
        """Get a pull request by number

        Args:
            repo_name (str): Name of the repository
            pr_number (int): Number of the pull request

        Returns:
            PullRequest: The pull request
        """
        repo: Repository = self.get_repository(repo_name)
        return repo.get_pull(pr_number)

    @action
    def comment_on_pull_request(
        self, repo_name: str, pr_number: int, body: str
    ) -> None:
        """Comment on a pull request

        Args:
            repo_name (str): Name of the repository
            pr_number (int): Number of the pull request
            body (str): Body of the comment
        """
        pr: PullRequest = self.get_pull_request(repo_name, pr_number)
        pr.create_issue_comment(body)

    @action
    def create_issue(
        self,
        repo_name: str,
        title: str,
        body: str = "",
        assignee: Optional[str] = None,
        labels: Optional[List[str]] = None,
    ) -> None:
        """Create an issue

        Args:
            repo_name (str): Name of the repository
            title (str): Title of the issue
            body (str, optional): Body of the issue. Defaults to "".
            assignee (str, optional): Username of the assignee. Defaults to None.
            labels (List[str], optional): List of label names. Defaults to None.
        """
        repo: Repository = self.get_repository(repo_name)
        repo.create_issue(title=title, body=body, assignee=assignee, labels=labels)  # type: ignore

    @observation
    def get_issue(self, repo_name: str, issue_number: int) -> Issue:
        """Get an issue by number

        Args:
            repo_name (str): Name of the repository
            issue_number (int): Number of the issue

        Returns:
            Issue: The issue
        """
        repo: Repository = self.get_repository(repo_name)
        return repo.get_issue(issue_number)

    @action
    def comment_on_issue(self, repo_name: str, issue_number: int, body: str) -> None:
        """Comment on an issue

        Args:
            repo_name (str): Name of the repository
            issue_number (int): Number of the issue
            body (str): Body of the comment
        """
        issue: Issue = self.get_issue(repo_name, issue_number)
        issue.create_comment(body)

    def _get_repository_contents_recursive(
        self, repo: Repository, path: str
    ) -> List[Dict[str, str]]:
        """Recursively get the contents of a repository

        Args:
            repo (Repository): The repository
            path (str): Path of the directory

        Returns:
            List[Dict[str, str]]: List of file information
        """
        contents = repo.get_contents(path)
        if not isinstance(contents, list):
            contents = [contents]

        files = []
        for content in contents:
            if content.type == "dir":
                files.extend(
                    self._get_repository_contents_recursive(repo, content.path)
                )
            else:
                files.append(
                    {"path": content.path, "name": content.name, "type": content.type}
                )
        return files

    @observation
    def list_repository_files(self, repo_name: str) -> List[Dict[str, str]]:
        """List all files in a repository

        Args:
            repo_name (str): Name of the repository

        Returns:
            List[Dict[str, str]]: List of file information
        """
        repo: Repository = self.get_repository(repo_name)
        return self._get_repository_contents_recursive(repo, "")

    @observation
    def get_repository_files(self, repo_name: str) -> Dict[str, str]:
        """Get all files in a repository

        Args:
            repo_name (str): Name of the repository

        Returns:
            Dict[str, str]: Dictionary of file paths and their contents
        """
        repo: Repository = self.get_repository(repo_name)
        files = self._get_repository_contents_recursive(repo, "")
        file_contents = {}
        for file in files:
            content = repo.get_contents(file["path"])
            if isinstance(content, list):
                continue  # Skip directories
            file_contents[file["path"]] = content.decoded_content.decode("utf-8")
        return file_contents

    @observation
    def get_repository_file(self, repo_name: str, path: str) -> str:
        """Get a single file from a repository

        Args:
            repo_name (str): Name of the repository
            path (str): Path of the file

        Returns:
            str: Content of the file
        """
        repo: Repository = self.get_repository(repo_name)
        content = repo.get_contents(path)
        if isinstance(content, list):
            raise ValueError(f"Path '{path}' points to a directory, not a file.")
        return content.decoded_content.decode("utf-8")
